from ..cyc import Organism
from ..cyc.BuildModel import BuildRawModelFromDB

class db(Organism):
    def __init__(self, path, *args, **kwargs):

        if '\\' in path:
            path.replace('\\','/')
        if path.endswith('/'):
            path = path[:-1]
        if '/' in path[:-1]:
            path, data = path.rsplit('/', 1)
            path += '/'
        else:
            data = str(path)
            path = ''

        Organism.__init__(self, path=path, data=data, *args, **kwargs)

    @property
    def reactions(self):
        return list(self.Reaction.values())

    @property
    def compounds(self):
        return list(self.Compound.values())

    @property
    def genes(self):
        return list(self.Gene.values())

    @property
    def proteins(self):
        return list(self.Protein.values())

    @property
    def enzrxns(self):
        return list(self.Enzrxn.values())

    @property
    def pathways(self):
        return list(self.Pathway.values())

    def Reactions(self, f=None):
        reacs = list(self.Reaction.keys())
        if f:
            for reac in reacs:
                if f not in reac:
                    reacs.remove(reac)
        return reacs

    def Compounds(self, f=None):
        cpds = list(self.Compound.keys())
        if f:
            for cpd in cpds:
                if f not in cpd:
                    cpds.remove(cpd)
        return cpds

    def Genes(self, f=None):
        gs = list(self.Gene.keys())
        if f:
            for g in gs:
                if f not in g:
                    gs.remove(g)
        return gs

    def Proteins(self, f=None):
        prots = list(self.Protein.keys())
        if f:
            for prot in prots:
                if f not in prot:
                    prots.remove(prot)
        return prots

    def Enzrnxs(self, f=None):
        ers = list(self.Enzrxn.keys())
        if f:
            for er in ers:
                if f not in er:
                    ers.remove(er)
        return ers

    def Pathways(self, f=None):
        ps = list(self.Pathway.keys())
        if f:
            for p in ps:
                if f not in p:
                    ps.remove(p)
        return ps

    def CompoundFormula(self, compound, IncCharge=True):
        formula = self.Compound[compound].EmpForm
        if IncCharge:
            charge = 0
            if self.Compound[compound].has_key('ATOM-CHARGES'):
                for a in self.Compound[compound]['ATOM-CHARGES']:
                    charge += int(a[1:-1].split(' ')[1])
            formula['Charge'] = charge
        return formula

    def CompoundsFormulae(self, IncCharge=True, f=None):
        rv = {}
        for c in self.Compounds(f=f):
            rv[c] = self.CompoundFormula(c, IncCharge=IncCharge)
        return rv

    def BuildModel(self):
        return BuildRawModelFromDB(self)

    def GroupReactions(self):
        rv = {}
        for r in self.reactions:
            if r.SubReactions():
                rv[r.UID] = r.SubReactions()
        return rv


import math
import numpy
import types
import multiprocessing
from ..classes.fva import fva
from .MinSolve import SetLinearMinFluxObjective
from cobra import Reaction
from cobra.flux_analysis import variability

def FVA(model, reaclist=None, subopt=1.0, IncZeroes=True, VaryOnly=False,
        AsMtx=False, tol=1e-10, PrintStatus=False, cobra=True, processes=None,
        loopless=False, pfba_factor=None, reset_state=True):
    if reset_state:
        state = model.GetState()
#    import time
#    print 'before fva solve ' + time.asctime(time.localtime(time.time()))
    model.Solve(PrintStatus=PrintStatus)
#    print 'after fva solve ' + time.asctime(time.localtime(time.time()))
    if model.solution.status != 'optimal' or math.isnan(model.solution.objective_value):
        statusmsg = model.solution.status
        model.SetState(state)
        raise ValueError("no optimal solution, problem " + statusmsg)
    else:
        if reaclist == None:
            reaclist = model.reactions
        elif isinstance(reaclist,str) or isinstance(reaclist, Reaction):
            reaclist = [reaclist]
        if cobra:
#            print 'before cobra fva ' + time.asctime(time.localtime(time.time()))
            reaclist = model.GetReactions(reaclist)
            fvadict = variability.flux_variability_analysis(model=model, 
                reaction_list=reaclist, fraction_of_optimum=subopt,
                loopless=loopless, pfba_factor=pfba_factor)
                #solver=model.solver, objective_sense=model.objective_direction)
#            print 'after cobra fva ' + time.asctime(time.localtime(time.time()))
            rv = fva({}, bounds=model.bounds)
#            for reac in fvadict:
#                lo = fvadict[reac]["minimum"] if abs(
#                        fvadict[reac]["minimum"]) > tol else 0.0
#                hi = fvadict[reac]["maximum"] if abs(
#                        fvadict[reac]["maximum"]) > tol else 0.0
#                rv[reac] = (lo,hi)
            for row in fvadict.iterrows():
                hi = row[1][0] if abs(row[1][0]) > tol else 0.0
                lo = row[1][1] if abs(row[1][1]) > tol else 0.0
                rv[model.GetReactionName(row[0])] = (lo,hi)
#                rv[model.GetReaction(row[0])] = (lo,hi)
        else:
#            print 'not cobra fva'
            model.DelSumReacsConstraint("FVA_objective")
            model.SetObjAsConstraint(name="FVA_objective", subopt=subopt)
            rv = fva({}, bounds=model.bounds)
            pool = multiprocessing.Pool(processes=processes)
            results = [pool.apply_async(FluxRange, args=(model, reac, tol,
                        False, True)) for reac in reaclist]
            pool.close()
            pool.join()
            for x in results:
                rv[model.GetReactionName(x.get().keys()[0])] = x.get().values()[0]
            if "FVA_objective_sum_reaction" in rv:
                rv.pop("FVA_objective_sum_reaction")
#            for reac in reaclist:
#                rv[str(reac)] = pool.apply_async(FluxRange, args=(model, reac, tol,
#                                                                False))
#            for reac in reaclist:
#                lo,hi = model.FluxRange(reac,tol=tol,reset_state=False)
#                if IncZeroes or abs(lo) > tol or abs(hi) > 0.0:
#                    rv[str(reac)] = (lo,hi)
            model.DelSumReacsConstraint("FVA_objective")
        if VaryOnly:
            rv = rv.Variable()
        if AsMtx:
            rv = rv.AsMtx()
#        print 'before fva set state ' + time.asctime(time.localtime(time.time()))
        if reset_state:
            model.SetState(state)
#        print 'after fva set state ' + time.asctime(time.localtime(time.time()))
        return rv


def MinFluxFVA(model, reaclist=None, subopt=1.0, IncZeroes=True, VaryOnly=False, 
                AsMtx=False, tol=1e-10, PrintStatus=False, cobra=True, 
                processes=None, weighting='uniform', ExcReacs=[],
                loopless=False, pfba_factor=1.0, reset_state=True):
    if reset_state:
        state = model.GetState()
    if (cobra) and (not ExcReacs) and (weighting == 'uniform'):
#        import time
#        print 'before cobra min flux fva ' + time.asctime(time.localtime(time.time()))
        rv = FVA(model, reaclist=reaclist, subopt=subopt, IncZeroes=IncZeroes, 
                VaryOnly=VaryOnly, AsMtx=AsMtx, tol=tol, 
                PrintStatus=PrintStatus, cobra=cobra, processes=processes,
                loopless=loopless, pfba_factor=pfba_factor, reset_state=False)
#        print 'after cobra min flux fva' + time.asctime(time.localtime(time.time()))
    else:
#        print 'not cobra min flux fva'
        model.SetObjAsConstraint('Primary_Objective_Constraint', subopt=subopt)
        model.SplitRev()
        SetLinearMinFluxObjective(model, weighting=weighting, ExcReacs=ExcReacs)

        if reaclist:
            tmp_reaclist = []
            for reac in reaclist:
                tmp_reaclist.append(model.GetReaction(reac))
            for reac in list(tmp_reaclist):
                if "reflection" in reac.notes:
                    if reac.reflection not in reaclist:
                        tmp_reaclist.append(reac.reflection)
            reaclist = model.GetReactionNames(tmp_reaclist)
        if not pfba_factor:
            pfba_factor = 1.0
        split_fva = FVA(model, reaclist=reaclist, subopt=pfba_factor, IncZeroes=IncZeroes, 
                    VaryOnly=VaryOnly, AsMtx=AsMtx, tol=tol, 
                    PrintStatus=PrintStatus, cobra=cobra, processes=processes)

        reverse_reactions = [x for x in model.reactions
                             if "reflection" in x.notes and
                             x.id.endswith('_reverse')]
        # If there are no reverse reactions, then there is nothing to do
        if not split_fva:
            model.MergeRev()
            model.DelObjAsConstraint('Primary_Objective_Constraint')
            model.SetState(state)
            return split_fva

        rv = fva(split_fva)
        for reac in rv.keys():
            cobra_reac = model.GetReaction(reac)
            if cobra_reac in reverse_reactions:
                reverse = cobra_reac
                forward = reverse.reflection
                lo_for, hi_for = rv[forward.id]
                lo_rev, hi_rev = rv[reverse.id]
                if ((lo_for > tol) and (hi_rev > tol)) or ((lo_rev > tol) and (hi_for > tol)):
                    raise ValueError(forward.id + " problem with FVA merging")
                if hi_rev < tol:
                    new_range = (lo_for, hi_for)
                elif hi_for <tol:
                    new_range = (-hi_rev, -lo_rev)
                else:
                    new_range = (-hi_rev, hi_for)
                rv[forward.id] = new_range
                rv.pop(reverse.id)
        model.MergeRev()
        model.DelObjAsConstraint('Primary_Objective_Constraint')
    for reac in rv.keys():
        if reac.endswith("_sum_reaction") or reac.endswith("_metbounds"):
            rv.pop(reac)
#    print 'before min flux fva set state ' + time.asctime(time.localtime(time.time()))
    if reset_state:
        model.SetState(state)
#    print 'after min flux fva set state ' + time.asctime(time.localtime(time.time()))
    return rv

def AllFluxRange(model, tol=1e-10, processes=None, reset_state=True):
    rangedict = fva({}, bounds=model.bounds)
    if reset_state:
        state = model.GetState()
    model.Solve(PrintStatus=False)
    if model.Optimal():
        pool = multiprocessing.Pool(processes=processes)
        results = [pool.apply_async(FluxRange, args=(model, reac.id, tol,
                        False, True)) for reac in model.reactions]
        pool.close()
        pool.join()
        for x in results:
            #print(type(x.get()))
            rangedict[list(x.get().keys())[0]] = list(x.get().values())[0]
#        for reac in model.reactions:
#            rangedict[reac.id] = pool.apply(FluxRange, args=(model, reac.id,
#                                                            tol, False))
#            rangedict[reac.id] = model.FluxRange(obj=reac.id, tol=tol,
#                                                reset_state=False)
    if reset_state:
        model.SetState(state)
    return rangedict

def FluxRange(model, obj, tol=1e-10, reset_state=True, return_reac=False):
    """ post: changes objective if resetobj = False!!! """
    if reset_state:
        state = model.GetState()
    model.ZeroObjective()
    model.SetObjective(obj)
    model.SetObjDirec("Min")
    model.Solve(PrintStatus=False)
    if model.solution.status != 'optimal':
        if model.solution.status == 'unbound':
            lo = float("-inf")
        else:
            lo = float("NaN")
    else:
        if math.isnan(model.solution.objective_value):
            lo = float("-inf")
        else:
            lo = model.GetObjVal()
    model.SetObjDirec("Max")
    model.Solve(PrintStatus=False)
    if model.solution.status != 'optimal':
        if model.solution.status == 'unbound':
            hi = float("inf")
        else:
            hi = float("NaN")
    else:
        if math.isnan(model.solution.objective_value):
            hi = model.float("inf")
        else:
            hi = model.GetObjVal()
    if abs(lo) < tol:
        lo = 0.0
    if abs(hi) < tol:
        hi = 0.0
    if reset_state:
        model.SetState(state)
    if return_reac:
        return {obj:(lo, hi)}
    else:
        return (lo, hi)

def FluxVariability(model, reffva=None, fva=None, excreacs=[], tol=1e-10,
                    getratio=False):
    if reffva == None:
        reffva = model.AllFluxRange(tol=tol)
    if fva == None:
        fva = model.FVA(tol=tol)
    rv = {}
    var = 0
    varyreac = 0
    for reac in reffva.keys():
        if reac not in excreacs:
            refrange = reffva[reac][1] - reffva[reac][0]
            if refrange > tol and (not numpy.allclose(
                                                refrange,float("inf"))):
                varyreac += 1
                reacrange = fva[reac][1] - fva[reac][0]
                ratio = reacrange/float(refrange)
                var += ratio
                rv[reac] = ratio
    avgvar = var/varyreac
    print ("total variability:", var)
    print ("number of variable reactions before optimisation:", varyreac)
    print ("averge variability:", avgvar)
    if getratio:
        return rv

def InternalCycles(model, allowedreacs=None, reacsbounds={}, tol=1e-10):
    """ pre: reacsbounds={reac:(lo,hi)}, all external reactions blocked
       post: model with reactions in internal cycles (for doing elementary modes) """
    state = model.GetState()
    model.SetConstraints(reacsbounds)
    if allowedreacs == None:
        rd = model.AllFluxRange()
        allowedreacs = rd.Allowed(tol=tol).keys()
    model.SetState(state)
    if len(allowedreacs) > 0:
        armodel = model.SubModel(allowedreacs)
        return armodel




def compareMetaboliteDicts(d1, d2): 
    sorted_d1_keys = sorted(d1.keys())
    sorted_d2_keys = sorted(d2.keys())
    for i in range(len(sorted_d1_keys)): 
        if not compareMetabolites(sorted_d1_keys[i], sorted_d2_keys[i], naive=True):
            return False
        elif not d1[sorted_d1_keys[i]] == d2[sorted_d2_keys[i]]: 
            return False
    else: 
        return True

def compareMetabolites(met1, met2, naive=False): 
    if isinstance(met1, set): 
        return compareReactions(list(met1), list(met2), naive)
    if isinstance(met1, list): 
        if not isinstance(met2, list): 
            return False
        elif len(met1) != len(met2):
            return False 
        else:
            for i in range(len(met1)): 
                if not compareMetabolites(met1[i], met2[i], naive): 
                    return False 
            else: 
                return True
    else: 
        if not True: 
            #can never be entered
            pass
        elif not met1._bound == met2._bound: 
            return False
        elif not met1._constraint_sense == met2._constraint_sense: 
            return False
        #elif not met1.annotation == met2.annotation: 
        #    return False
        elif not met1.charge == met2.charge: 
            return False
        elif not met1.compartment == met2.compartment: 
            return False
        elif not met1.name == met2.name: 
            return False
        elif not met1.compartment == met2.compartment: 
            return False
        #elif not met1.notes == met2.notes: 
        #    return False
        elif not naive:
            if not compareReactions(met1._reaction, met2._reaction, naive=True): 
                return False
            elif not compareModels(met1._model, met2._model, naive=True): 
                return False
            else: 
                return True
        else: 
            return True

def compareReactions(r1, r2, naive=False): 
    if isinstance(r1, set): 
        return compareReactions(list(r1), list(r2), naive)
    if isinstance(r1, list): 
        if not isinstance(r2, list): 
            return False
        elif len(r1) != len(r2):
            return False 
        else:
            for i in range(len(r1)): 
                if not compareReactions(r1[i], r2[i],naive): 
                    return False 
            else: 
                return True
    else: 
        if not True: 
            #can never be entered
            pass
        #elif not r1._compartments == r2._compartments: 
        #    return False
        #elif not r1._forward_variable == r2._forward_variable:
        #    return False
        elif not r1._gene_reaction_rule == r2._gene_reaction_rule:
            return False
        elif not r1._id == r2._id: 
            return False
        elif not r1._lower_bound == r2._lower_bound: 
            return False 
        #elif not r1._model == r2._model: 
        #    return False
        #elif not r1._reverse_variable == r2._reverse_variable: 
        #    return False 
        elif not r1._upper_bound == r2._upper_bound: 
            return False 
        #elif not r1.annotation == r2.annotation: 
        #    return False 
        elif not r1.name== r2.name: 
            return False
        #elif not r1.notes == r2.notes: 
        #    return False 
        elif not r1.subsystem == r2.subsystem: 
            return False 
        elif not r1.variable_kind == r2.variable_kind: 
            return False 
        elif not naive:
            if not compareMetaboliteDicts(r1._metabolites, r2._metabolites): 
                return False 
            elif not compareGenes(r1._genes,r2._genes, naive=True): 
                return False
            else: 
                return True
        else: 
            return True
   

def compareGenes(g1, g2, naive=False): 
    if isinstance(g1, set): 
        return compareGenes(list(g1), list(g2), naive)
    if isinstance(g1, list): 
        if not isinstance(g2, list): 
            return False
        elif len(g1) != len(g2):
            return False 
        else:
            for i in range(len(g1)): 
                if not compareGenes(g1[i], g2[i], naive): 
                    return False 
            else: 
                return True
    else: 
        if not True: 
            #can never be entered
            pass
        elif not g1._functional == g2._functional: 
            return False
        elif not g1._id == g2._id:
            return False
        #elif not g1._model == g2._model:
        #    return False
        elif not g1.annotation == g2.annotation: 
            return False
        elif not g1.name == g2.name: 
            return False 
        #elif not g1.notes == g2.notes: 
        #    return False
        elif not naive: 
            if not compareReactions(g1._reaction,g2._reaction, naive=True): 
                return False
            else:
                return True
        else: 
            return True
    

def compareModels(m1, m2, naive=False): 
    if not True: 
        #can never be entered
        pass
    #elif not m1._compartments == m2._compartments: 
    #    return False
    #elif not m1._contexts == m2._contexts:
    #    return False
    #elif not m1._solver == m2._solver:
    #    return False
    elif not m1._id == m2._id: 
        return False
    #elif not m1._trimmed == m2.trimmed: 
    #    return False
    #elif not m1._trimmed_genes == m2._trimmed_genes: 
    #    return False 
    #elif not m1._trimmed_reactions == m2._trimmed_reactions: 
    #    return False 
    #elif not m1.annotation == m2.annotation: 
    #    return False 
    elif not m1.bounds == m2.bounds: 
        return False 
    elif not m1.name == m2.name: 
        return False
    #elif not m1.notes == m2.notes: 
    #    return False 
    #elif not m1.quadratic_component == m2.quadratic_component: 
    #    return False
    elif not naive: 
        if not compareGenes(m1.genes, m2.genes): 
            return False 
        elif not compareMetabolites(m1.metabolites, m2.metabolites): 
            return False 
        elif not compareReactions(m1.reactions,m2.reactions): 
            return False
        else: 
            return True
    else: 
        return True



#import pandas
from .matrix import matrix

class flux(dict):

    def __init__(self, *args, **kwargs):
        # if type(self) == pandas.core.series.Series:
        #     fd = {}
        #     for r in self.axes[0]:
        #         fd[r] = self[r]
        super(flux, self).__init__(*args, **kwargs)

    def __call__(self, string=''):
        rv = flux()
        for r in self:
            if string in r:
                rv[r] = self[r]
        return rv

    def Filter(self, lo=0, hi=float('inf'), f=None, abs=True):
        sol = self.Copy()
        if isinstance(lo, str):
            f = lo
            lo = 0
        if f != None:
            for reac in list(sol.keys()):
                if f not in reac:
                    del sol[reac]
        for reac in list(sol.keys()):
            if abs:
                if not (abs(sol[reac]) >= lo and abs(sol[reac]) <= hi):
                    del sol[reac]
            else:
                if not (sol[reac] >= lo and sol[reac] <= hi):
                    del sol[reac]
        return sol


    def Print(self, lo=0, hi=float('inf'), f=None, Sort="value",
              sortabs=True, reverse=True):
        sol = self.Copy()
#        sol.Filter(lo=lo, hi=hi, f=f)
        if isinstance(lo, str):
            f = lo
            lo = 0
        if f != None:
            for reac in list(sol.keys()):
                if f not in reac:
                    del sol[reac]
        if Sort == "value":
            if sortabs:
                function = lambda k:abs(k[1])
            else:
                function = lambda k:k[1]
            for key,value in sorted(sol.items(), key=function,reverse=reverse):
                if abs(value) >= lo and abs(value) <= hi:
                    print("%s: %s" % (key, value))
        elif Sort == "key":
            for key in sorted(iter(sol.keys())):
                if abs(sol[key]) >= lo and abs(sol[key]) <= hi:
                    print("%s: %s" % (key, sol[key]))

    def Diff(self, fd, IncZeroes=False, AsMtx=False, tol=1e-10):
        rv = {}
        sol2only = list(set(fd.keys()).difference(self.keys()))
        for reac in self.keys():
            if reac in fd.keys():
                rv[reac] = self[reac] - fd[reac]
            else:
                rv[reac] = self[reac]
        for reac in sol2only:
            rv[reac] = -fd[reac]
        if not IncZeroes:
            for reac in list(rv.keys()):
                if abs(rv[reac]) < tol:
                    del rv[reac]
        if AsMtx:
            mtx = matrix({"diff":rv})
            rv = mtx
        return rv

    def AsMtx(self):
        return matrix({"Flux":self})

    def Copy(self):
        return flux(self)


import numpy
from .matrix import matrix

class fva(dict):

    def __init__(self, content, bounds=float("inf")):
        super(fva, self).__init__(content)
        self.bounds = bounds

    def __call__(self, string=''):
        rv = fva({})
        for r in self:
            if string in r:
                rv[r] = self[r]
        return rv

    def Blocked(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            if (abs(self[reac][0]) < tol) and (abs(self[reac][1]) < tol):
                rv[reac] = self[reac]
        return rv

    def Allowed(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            if (abs(self[reac][0]) > tol) or (abs(self[reac][1]) > tol):
                rv[reac] = self[reac]
        return rv

    def Variable(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            diff = self[reac][0] - self[reac][1]
            if abs(diff) > tol:
                rv[reac] = self[reac]
        return rv

    def Fixed(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            diff = self[reac][0] - self[reac][1]
            if abs(diff) < tol:
                rv[reac] = self[reac]
        return rv

    def Essential(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            if (self[reac][0] > tol) or (self[reac][1] < -tol):
                rv[reac] = self[reac]
        return rv

    def Substitutable(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            diff = self[reac][0] - self[reac][1]
            if (self[reac][0] <= tol and self[reac][1] >= -tol) and abs(diff) > tol:
                rv[reac] = self[reac]
        return rv

    def Unbounded(self):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
#            if self[reac][0] == float("-inf") or self[reac][1] == float("inf"):
            if numpy.allclose(self[reac][0],-self.bounds) or numpy.allclose(self[reac][1],self.bounds):
                rv[reac] = self[reac]
        return rv

    def Bounded(self):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
#            if self[reac][0] != float("-inf") and self[reac][1] != float("inf"):
            if (not numpy.allclose(self[reac][0],-self.bounds)) and (not numpy.allclose(self[reac][1],self.bounds)):
                rv[reac] = self[reac]
        return rv

    def BothDirections(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            if self[reac][0] < -tol and self[reac][1] > tol:
                rv[reac] = self[reac]
        return rv

    def ForwardOnly(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            if self[reac][0] > -tol and self[reac][1] > tol:
                rv[reac] = self[reac]
        return rv

    def BackwardOnly(self,tol=1e-10):
        rv = fva({}, bounds=self.bounds)
        for reac in self:
            if self[reac][0] < -tol and self[reac][1] < tol:
                rv[reac] = self[reac]
        return rv

    def AsMtx(self):
        return matrix(self, index=['Min','Max']).transpose()

    def GetReacs(self,reacs):
        rv = fva({}, bounds=self.bounds)
        for reac in reacs:
            if reac in self.keys():
                rv[reac] = self[reac]
        return rv

    def GetReacsMeanFlux(self,reacs):
        rv = {}
        for reac in reacs:
            rv[reac] = (self[reac][0]+self[reac][1])/2.0
        return rv

    def MaxDiff(self):
        diff = []
        for reac in self.keys():
            diff.append(abs(self[reac][1]-self[reac][0]))
        return max(diff)

    def FluxRangeDiff(self, fd, thres=1):
        """ a measure of flux difference with another fva object
            pre: thres -- only get reactions with difference > thres """
        rv = {}
        both1n2 = list(set(self.keys()).intersection(fd.keys()))
        diffdirec = self.FluxDiffDirec(fd) # reac in different direc
        comparereacs = list(set(both1n2).difference(diffdirec))
        overlaps = self.FluxRangeOverlap(fd)   # reac with overlap flux range
        comparereacs = list(set(comparereacs).difference(overlaps))
        for r in comparereacs:
            max1 = max(self[r])
            min1 = min(self[r])
            max2 = max(fd[r])
            min2 = min(fd[r])
            if (max1 > 0) or (max2 > 0):    # positive flux
                if (min1*thres) > max2:
                    rv[r] = max2/min1
                elif (min2*thres) > max1:
                    rv[r] = -max1/min2
            elif (min1 < 0) or (min2 < 0):  # negative flux
                if (max1*thres) < min2:
                    rv[r] = min2/max1
                elif (max2*thres) < min1:
                    rv[r] = -min1/max2
            else:
                print(r)
        return rv

    def FluxDiffDirec(self, fd):
        """ Compare with another fva object and get reactions with different directions """
        rv = []
        both1n2 = list(set(self.keys()).intersection(fd.keys()))
        for r in both1n2:
            max1 = max(self[r])
            min1 = min(self[r])
            max2 = max(fd[r])
            min2 = min(fd[r])
            if (min1 >= 0) and (max2 <= 0):
                rv.append(r)
            elif (min2 >= 0) and (max1 <= 0):
                rv.append(r)
        return rv

    def FluxRangeOverlap(self, fd):
        """ Compare with another fva object and get reactions with overlapping flux range """
        both1n2 = list(set(self.keys()).intersection(fd.keys()))
        nonoverlap = []
        for r in both1n2:
            max1 = max(self[r])
            min1 = min(self[r])
            max2 = max(fd[r])
            min2 = min(fd[r])
            if (min1 > max2) or (min2 > max1):
                nonoverlap.append(r)
        rv = list(set(both1n2).difference(nonoverlap))
        return rv



""" Not functional! To be modified """

def MatchFlux(self, md, vd, lo=0, hi=100, samedirec=True, count=50, 
                display=False, tol=1e-6):
    """ pre: md = matched flux dict; vd = vary flux dict / vary met bound dict """
    state = self.GetState()
    avg = (float(lo)+float(hi))/2
    if set(vd.keys()).issubset(self.cnames.values()):    # vary flux dict
        self.SetSumReacsConstraint(vd,avg)
    elif set(vd.keys()).issubset(self.rnames.values()) or set(vd.keys()).issubset(self.smexterns.rnames): # vary met bound dict
        metdic = {}
        for met in vd:
            metdic[met] = (avg*vd[met],avg*vd[met])
        self.SetMetBounds(metdic)
    else:
        print("vary dict error")
    self.Solve(False)
    reac = md.keys()[0]
    val = md[reac]
    sol = self.GetSol(IncZeroes=True)[reac]
    if set(vd.keys()).issubset(self.cnames.values()):    # vary flux dict
        self.DelSumReacsConstraint()
    elif set(vd.keys()).issubset(self.rnames.values()) or set(vd.keys()).issubset(self.smexterns.rnames): # vary met bound dict
        for met in vd:
            metdic[met] = (0,0)
        self.SetMetBounds(metdic)   # set met bounds back to balance
    self.SetState(state)
    if display:
        print("Varying value = " + str(avg))
        print(reac + " " + str(sol))
    if abs(1-(sol/val)) < tol:
        return avg
    else:
        if samedirec:
            if sol > val:
                hi = avg
            else:
                lo = avg
        else:
            if sol > val:
                lo = avg
            else:
                hi = avg
        count = count - 1
        if count > 0:
            return self.MatchFlux(md, vd, lo, hi, samedirec, count, display, tol)
        else:
            return None
            print("Varying  value = " + str(avg))
            print(reac + " " + str(sol))

def MatchRatio(self,numdic,domdic,val,vd,lo=0,hi=100,samedirec=True,count=50,display=False,tol=1e-6):
    state = self.GetState()

    avg = (float(lo)+float(hi))/2
    if set(vd.keys()).issubset(self.cnames.values()):    # vary flux dict
        self.SetSumReacsConstraint(vd,avg,'matchratio')
    elif set(vd.keys()).issubset(self.rnames.values()) or set(vd.keys()).issubset(self.smexterns.rnames): # vary met bound dict
        metdic = {}
        for met in vd:
            metdic[met] = (avg*vd[met],avg*vd[met])
        self.SetMetBounds(metdic)
    else:
        print("vary dict error")

    self.Solve(False)
    num = 0.0
    for reac in numdic:
        num += self.GetSol(IncZeroes=True)[reac]*numdic[reac]
    dom = 0.0
    for reac in domdic:
        dom += self.GetSol(IncZeroes=True)[reac]*domdic[reac]
    sol = num/dom
    self.SetState(state)
    if set(vd.keys()).issubset(self.cnames.values()):    # vary flux dict
        self.DelSumReacsConstraint('matchratio')
    elif set(vd.keys()).issubset(self.rnames.values()) or set(vd.keys()).issubset(self.smexterns.rnames): # vary met bound dict
        for met in vd:
            metdic[met] = (0,0)
        self.SetMetBounds(metdic)   # set met bounds back to balance

    if display:
        print("Varying  value = " + str(avg))
        print(reac + " " + str(sol))
    if abs(1-(sol/val)) < tol:
        self.SetState(state)
        return avg
    else:
        if samedirec:
            if sol > val:
                hi = avg
            else:
                lo = avg
        else:
            if sol > val:
                lo = avg
            else:
                hi = avg
        count = count - 1
        if count > 0:
            return self.MatchRatio(numdic, domdic, val, vd, lo, hi, samedirec, count, display, tol)
        else:
            self.SetState(state)
            print("Varying  value = " + str(avg))
            print("Ratio =",sol)



import builtins as exceptions
#many unsupported attribute of types lib in python 3
import types
import re, math 
from collections import defaultdict
import numpy
import scipy

import cobra
from cobra import Metabolite, Reaction, Gene
from cobra.flux_analysis import deletion, moma, phenotype_phase_plane
from cobra.core.solution import get_solution
#from cobra.manipulation import modify

#from ..analysis import FCA, Pareto, RWFM, MOMA, ROOM, GeometricFBA, MinSolve
from ..analysis import Graph, FluxSum, FVA, MinSolve, Scan
from ..manipulation import Reversible
from ..classes.flux import flux
from ..io import Network

#############################################################################


class model(cobra.Model):
    def __init__(self, existing_model=None, bounds=1000.0):
        if type(existing_model) == model:
            self.__dict__ = existing_model.__dict__
        else:
            cobra.Model.__init__(self, existing_model)
            self.objective_direction = "minimize"
            #self.solver = None
            self.quadratic_component = None
            self.bounds = bounds
            self.SetBounds(bounds=bounds)

    #### MANIPULATING AND WRITING MODELS #########################################

    def Copy(self):
        return model(self.copy())

    def SubModel(self, reactions):
        new_model = model()
        for reac in reactions:
            if not isinstance(reac, Reaction):
                reac = self.GetReaction(reac)
            new_model.add_reaction(reac)
        return new_model

    def DuplicateModel(self, suffixes):
        """ suffixes = list of strings of suffixes """
        big_model = model()
        for sf in suffixes:
            sf_model = self.copy()
            for reac in sf_model.reactions:
                reac.id += sf
            for met in sf_model.metabolites:
                met.id += sf
                if isinstance(met.compartment, str): 
                    met.compartment += sf
            sf_model.repair()
            big_model.MergeWithModel(sf_model)
        return big_model

    def MergeWithModel(self, other_model, replace_with_new=False):
        """ keep attributes of current model if there is repetition in IDs  """
        for reac in other_model.reactions:
            if reac.id not in self.Reactions():
                self.add_reaction(reac)
            else:
                if replace_with_new:
                    self.DelReaction(reac.id)
                    self.add_reaction(reac)  
                      
#    def WriteModel(self, filename, model_format=None, excel_format="cobra",
#                   sbml_level=2, sbml_version=1, fbc=False, ExtReacs=[]):
#        """ model_format = "sbml" | "excel" | "matlab" | "json" | "cobra" | "cobra_old" | "scrumpy" """
    def WriteModel(self, filename, model_format=None, excel_format="cobra", ExtReacs=[], **kwargs):
        """ model_format = "sbml" | "sbml_legacy" | "excel" | "matlab" | "json" | "cobra" | "cobra_old" | "scrumpy" | "yaml" """
        if self.id == None:
            self.id = 'None'
        from ..io import IO
        IO.WriteModel(model=self, filename=filename, model_format=model_format,
                      excel_format=excel_format, ExtReacs=ExtReacs, **kwargs)
#        IO.WriteModel(model=self, filename=filename, model_format=model_format,
#            excel_format=excel_format, sbml_level=sbml_level,
#            sbml_version=sbml_version, fbc=fbc, ExtReacs=ExtReacs)

    def WriteFile(self, *args, **kwargs):
        self.WriteModel(*args, **kwargs)

    def ToFile(self, *args, **kwargs):
        self.WriteModel(*args, **kwargs)

    ### SPLITTING AND MERGING REVERSIBLE REACTIONS ###############################################

    def SplitRev(self):
        Reversible.SplitRev(self)

    def MergeRev(self, update_solution=True):
        Reversible.MergeRev(self, update_solution=update_solution)

    ### WRITING NETWORKS AND ATTRIBUTES ##########################################################
    def WriteNetwork(self, filename, network_type="rm", ExtReacs=[], ExtMets=[]):
        """ network_type = "rr" | "mr" | "rm" | "mm" """
        ExtReacs = self.GetReactionNames(ExtReacs)
        ExtMets = self.GetMetaboliteNames(ExtMets)
        if network_type == "rr":
            Network.WriteReactionsToReactionsNetwork(self, filename,
                                    ExtReacs=ExtReacs, ExtMets=ExtMets)
        elif network_type == "mr" or network_type == "rm":
            Network.WriteReactionsToMetabolitesNetwork(self, filename,
                                    ExtReacs=ExtReacs, ExtMets=ExtMets)
        elif network_type == "mm":
            Network.WriteMetabolitesToMetabolitesNetwork(self, filename,
                                    ExtReacs=ExtReacs, ExtMets=ExtMets)

    def WriteAttributes(self, filename, attributes=[], node_type="reactions"):
        """ objects = "reactions" | "metabolites" """
        if 'reac' in node_type:
            Network.WriteReactionsAttributes(self, filename, attributes)
        elif 'met' in node_type:
            Network.WriteMetabolitesAttributes(self, filename, attributes)

    #### REACTIONS, METABOLITES AND GENE DATA ######################
    
    ######## GETTING REACTIONS #############################################
    def GetReaction(self, reac):
        if not isinstance(reac, Reaction):
            reac = self.reactions[self.reactions.index(reac)]
        return reac

    def GetReactions(self, reactions):
        return [self.GetReaction(reac) for reac in reactions]

    def GetReactionName(self, reac):
        if isinstance(reac, Reaction):
            reac = reac.id
        return reac

    def GetReactionNames(self, reactions):
        return [self.GetReactionName(reac) for reac in reactions]

    def Reactions(self, f=None):
        reacs = self.reactions.list_attr("id")
        if f:
            for reac in list(reacs):
                if f not in reac:
                    reacs.remove(reac)
        return reacs

    def Isozymes(self):
        """
            This function returns a list of lists of isozymes
        """
        iso = []
        skip = []
        for x in range(len(self.reactions)):
            if x not in skip:
                iso_x = [self.reactions[x].id]
                for y in range(x+1, len(self.reactions)):
                    if self.reactions[x].metabolites == self.reactions[y].metabolites:
                        iso_x.append(self.reactions[y].id)
                        skip.append(y)
                if len(iso_x) > 1:
                    iso.append(iso_x)
        return iso
        

    ######## PRINTING REACTIONS ######################################################
    def PrintReaction(self, reaction, AsMetNames=False):
        reacname = self.GetReactionName(reaction)
        reacstoi = self.GetReaction(reaction).build_reaction_string(AsMetNames)
        print(reacname + '\t' + reacstoi)

    def PrintReactions(self, reactions=None, AsMetNames=False):
        if reactions == None:
            reactions = self.reactions
        elif isinstance(reactions,str):
            reactions = self.Reactions(reactions)
        for reac in reactions:
            self.PrintReaction(reac, AsMetNames=AsMetNames)


    ######## ADDING AND REMOVING REACTIONS ###########################
    def AddReaction(self, reac, stodic, rev=False, bounds=None, name=None,
                    subsystem=None):
        """ bounds = val | (lb,ub) """
        reaction = Reaction(reac)
        if name != None:
            reaction.name = name
        if subsystem != None:
            reaction.subsystem = subsystem
        if bounds != None:
            if isinstance(bounds, tuple) or isinstance(bounds, list):
                reaction.lower_bound = bounds[0]
                reaction.upper_bound = bounds[1]
            else:
                reaction.lower_bound = bounds
                reaction.upper_bound = bounds
        else:
            reaction.upper_bound = self.bounds
            if rev:
                reaction.lower_bound = -self.bounds
            else:
                reaction.lower_bound = 0.0
        newstodic = {}
        for met in stodic:
            stoi = stodic[met]
            if not isinstance(met, Metabolite):
                if met not in self.Metabolites():
                    met = Metabolite(id=met)
                else:
                    met = self.metabolites.get_by_id(met)
            newstodic[met] = stoi
        reaction.add_metabolites(newstodic)
        self.add_reaction(reaction)

    #BUGGED DOES NOT DELETE THE REACTANT ADDED--> CAUSE BUG WITH: DEADENDMETABOLITES, PERIPHERALMETABOLITES():"Produced","Consumed",""
    def DelReaction(self, reaction, delete_metabolites=False):
        reaction = self.GetReaction(reaction)
        if delete_metabolites:
            for met in reaction.metabolites: 
                self.DelMetabolite(met)
        self.remove_reactions([reaction])

    def DelReactions(self, reactions, delete_metabolites=False):
        """ reactions = list of reactions """
        for reac in reactions:
            self.DelReaction(reac, delete_metabolites=delete_metabolites)
        #self.remove_reactions(reactions)

    def ChangeReactionStoichiometry(self, reaction, metstoidic, combine=False):
        reaction = self.GetReaction(reaction)
        metabolites = {}
        for met_name in metstoidic:
            met = self.GetMetabolite(met_name)
            metabolites[met] = metstoidic[met_name]
        reaction.add_metabolites(metabolites, combine=combine)

    ######## CHECKING IMBALANCES ###########################################
    def ImbalanceReactions(self, elements=None, IncCharge=True, ExcReacs=None,
                           ExcElements=None):
        rv = {}
        reactions = self.reactions
        if ExcReacs:
            ExcReacs = self.GetReactions(ExcReacs)
            reactions = list(set(reactions).difference(ExcReacs))
        for reac in reactions:
            bal_dict = self.CheckReactionBalance(reac, IncCharge=IncCharge,
                                                 ExcElements=ExcElements)
            if bal_dict:
                if not elements:
                    rv[reac.id] = bal_dict
                elif isinstance(elements,str):
                    if elements in bal_dict:
                        rv[reac.id] = bal_dict
                elif isinstance(elements,list):
                    if set(elements).intersection(bal_dict.keys()):
                        rv[reac.id] = bal_dict
        return rv


    def CheckReactionBalance(self, reac, IncCharge=True, ExcElements=None):
        reac = self.GetReaction(reac)
        reaction_element_dict = defaultdict(list)
        for the_metabolite, the_coefficient in reac._metabolites.items():
            if the_metabolite.elements is not None:
                [reaction_element_dict[k].append(the_coefficient*v)
                for k, v in the_metabolite.elements.items()]
                if ExcElements:
                    if isinstance(ExcElements,str):
                        ExcElements = [ExcElements]
                    if len(set(ExcElements).intersection(
                        reaction_element_dict.keys())) > 0:
                        return {}
            if (the_metabolite.charge is not None) and IncCharge:
                reaction_element_dict['Charge'].append(the_metabolite.charge
                                                        *the_coefficient)
        reaction_element_dict = dict([(k, sum(v))
                                for k, v in reaction_element_dict.items()])
        for element in list(reaction_element_dict.keys()):
            if numpy.allclose(reaction_element_dict[element], 0):
                del reaction_element_dict[element]
            elif isinstance(reaction_element_dict[element],float):
                if reaction_element_dict[element].is_integer():
                    reaction_element_dict[element] = int(
                                            reaction_element_dict[element])
#        if sum(map(abs, reaction_element_dict.values())) != 0:
#            return [reac.id, reaction_element_dict]
#        else:
        return dict(reaction_element_dict)


    ######## GETTING METABOLITES #############################################
    def GetMetabolite(self, met):
        if not isinstance(met, Metabolite):
            met = self.metabolites[self.metabolites.index(met)]#why create a new metabolite?
        return met

    def GetMetabolites(self, metabolites):
        return [self.GetMetabolite(met) for met in metabolites]

    def GetMetaboliteName(self, met):
        if isinstance(met, Metabolite):
            met = met.id
        return met

    def GetMetaboliteNames(self, metabolites):
        return [self.GetMetaboliteName(met) for met in metabolites]

    def Metabolites(self, f=None):
        mets = self.metabolites.list_attr("id")
        if f:
            for met in list(mets):
                if f not in met:
                    mets.remove(met)
        return mets

    ####### ADDING AND REMOVING METABOLITES #############################
    def AddMetabolite(self, met, formula=None, name=None, charge=None, compartment=None):
        if met in self.Metabolites():
            print(met + " is already in the model")
        else:
            metabolite = Metabolite(id=met, formula=formula, name=name,
                                charge=charge,compartment=compartment)
            self.add_metabolites([metabolite])

    def DelMetabolite(self, met, destructive=False, method='substractive'):
        """ method = 'subtractive'|'destructive' """
        met = self.GetMetabolite(met)
        if method == 'substractive': 
            destructive = False
        #if method == 'destructive':
        #    for reac in list(met._reaction):
                #reac.remove_from_model()
        #        self.DelReaction(reac)
        #met.remove_from_model(method=method)   
        met.remove_from_model(destructive=destructive)

    def DelMetabolites(self, mets, method='subtractive'):
        for met in mets:
            self.DelMetabolite(met, method=method)

    def SubstituteMetabolite(self, met_from, met_to):
        met_from = self.GetMetabolite(met_from)
        met_to = self.GetMetabolite(met_to)
        iw = self.InvolvedWith(met_from)
        for r in iw:
            r.add_metabolites({met_to:iw[r]}, combine=True)
            r.add_metabolites({met_from:-iw[r]}, combine=True)

    def AddProtonsToMets(self,met_proton_dic,proton,ExcReacs=None):
            self.AddProtonsToMet(met,proton,met_proton_dic[met],ExcReacs=ExcReacs)

    def AddProtonsToMet(self,met,proton,n_p,ExcReacs=None):
        """
            This function adds n_p amount of protons to the reactions met is involved in 
        """
        proton = self.GetMetabolite(proton)
        reactions = self.InvolvedWith(met,'metabolite')
        if ExcReacs:
            if isinstance(ExcReacs,str):
                ExcReacs = [ExcReacs]
            ExcReacs = self.GetReactions(ExcReacs)
            for excreac in ExcReacs:
                if excreac in reactions:
                    del reactions[excreac]
        for reac in reactions:
            reac.add_metabolites({proton:reactions[reac]*-n_p},combine=True)

    def AssignMetabolitesNeutralFormula(self):
        element_re = re.compile("([A-Z][a-z]?)([0-9.]+[0-9.]?|(?=[A-Z])?)")
        for met in self.metabolites:
            if met.charge != None and met.formula != None:
                neutral_formula = ''
                tmp_formula = met.formula.replace("*", "")
                parsed = element_re.findall(tmp_formula)
                for (element, count) in parsed:
                    if element != "H":
                        neutral_formula += element + str(count)
                    else:   # H not added if H is not in the formula
                        if count == '':
                            count = 1
                        count = float(count)
                        if count.is_integer():
                            count = int(count)
                        count -= met.charge
                        if count == 1:
                            neutral_formula += element
                        elif count != 0:
                            neutral_formula += element + str(count)
                        else:
                            if len(parsed) == 1:    # for proton
                                neutral_formula += "H"
                met.neutral_formula = neutral_formula

    ######## GETTING GENES ############################################
    def GetGene(self, gene):
        if not isinstance(gene, Gene):
            gene = self.genes[self.genes.index(gene)]
        return gene

    def GetGenes(self, genes):
        return [self.GetGene(gene) for gene in genes]

    #Fixed Bug with passing random String
    def GetGeneName(self, gene):
        if isinstance(gene, Gene):
            gene = gene.id
        return gene

    def GetGeneNames(self, genes):
        return [self.GetGeneName(gene) for gene in genes]

    def Genes(self, f=None):
        gs = self.genes.list_attr("id")
        if f:
            for g in list(gs):
                if f not in g:
                    gs.remove(g)
        return gs

    ######### GENE DELETIONS #######################################################
    def SingleDeletion(self, element_list=None, method='fba', element_type='gene', solver=None):
        return deletion.single_gene_deletion(self, method=method)

    def EssentialGenes(self,tol=1e-10):
        sdel = dict(deletion.single_gene_deletion(self))
        rv = []
        for i in range(0,len(sdel['status'])):
            if sdel['status'][i] == "infeasible":
                rv.append(list(dict(sdel['status']).keys()[i])[0])
        return rv

    def EssentialReactions(self,tol=1e-10):
        sdel = dict(deletion.single_reaction_deletion(self))
        rv = []
        for i in range(0,len(sdel['status'])):
            if sdel['status'][i] == "infeasible":
                rv.append(list(dict(sdel['status']).keys()[i])[0])
        return rv

    def DoubleDeletion(self,element_list_1=None, element_list_2=None, method='fba', single_deletion_growth_dict=None, element_type='gene', solver=None, number_of_processes=None, return_frame=True, zero_cutoff=1e-12, **kwargs):
        """ NOTE: bug with negative value for gene deletion """
        if element_type == "reaction":
            return deletion.double_reaction_deletion(self, element_list_1,
                                            element_list_2, **kwargs)
        elif element_type == "gene":
            return deletion.double_gene_deletion(self, element_list_1,
                                        element_list_2, **kwargs)
        else:
            raise Exception("unknown element type")

        #return double_deletion(self, element_list_1=element_list_1, element_list_2=element_list_2, method=method, single_deletion_growth_dict=single_deletion_growth_dict, element_type=element_type, solver=solver, number_of_processes=number_of_processes, return_frame=return_frame, zero_cutoff=zero_cutoff, **kwargs)

    #### CLEANING FUNCTIONS ########################################
    def Clean(self,reac=True,met=True,gene=True):
        if(met):
            in_use_mets=self.CollateActiveMetabolites();
            dead_end_mets = self.DeadEndMetabolites();
            delete_ = []
            for m in dead_end_mets:
                if in_use_mets.get(m) is None:
                    delete_.append(m)
            
            self.DelMetabolites(delete_)


    def CollateActiveMetabolites(self, reactions=None, AsMetNames=False):
        in_use_mets={}
        if reactions == None:
            reactions = self.reactions
        elif isinstance(reactions,str):
            reactions = self.Reactions(reactions)
        for reac in reactions:
            r = self.GetReaction(reac)
            for m in r.reactants:
                in_use_mets[m._id]=1
            for m in r.products:
                in_use_mets[m._id]=1
        return in_use_mets

    #### ASSOCIATIONS BETWEEN ATTRIBUTES #####################################
    
    def InvolvedWith(self, thing, thing_type=None, AsName=False):
        """ thing_type = None | "reaction" | "metabolite" """
        """
            This functions either 
                1) takes in a reaction object and returns a dict of {<metabolites involved in it> : <stoiciometry> } 
            or  2) takes in a metabolite object and returns a dict of {<reactions it is involved in> : stoichiometry} 
        """
        if thing in self.Reactions() or thing in self.reactions:
            if thing_type == None or 'reac' in thing_type:
                thing = self.GetReaction(thing)
                rv = dict(thing.metabolites)
                if AsName:
                    for m in list(rv.keys()):
                        rv[self.GetMetaboliteName(m)] = rv.pop(m)
                return rv
        elif thing in self.Metabolites() or thing in self.metabolites:
            if thing_type == None or 'met' in thing_type:
                thing = self.GetMetabolite(thing)
                reactions = thing.reactions
                rv = {}
                for reac in reactions:
                    if AsName:
                        rv[self.GetReactionName(reac)
                            ] = reac.metabolites[thing]
                    else:
                        rv[reac] = reac.metabolites[thing]
                return rv

    def DictConversion(self, input_dict=None, reaction_dict=None,metabolite_dict=None):
        #TAKES IN AN OBJECT DICTIONARY AND RETURN A NEW DICTIONARY WITH THE ID AS KEY AND OBJECT AS VALUE
        
        if(input_dict and not reaction_dict and not metabolite_dict):
            if(isinstance(list(input_dict.keys())[0],cobra.Metabolite)):
                metabolite_dict = input_dict
                #print("1")
            elif(isinstance(list(input_dict.keys())[0],cobra.Reaction)):
                reaction_dict = input_dict
                #print("2")
            elif(hasattr(list(input_dict.keys())[0],'id')):
                input_dict
            else:
                raise Exception("Bad input")
        """
        if(reaction_dict):
            #print("2")
            result = {}
            for reac in reaction_dict:
                result[reac.id] = reac
            return result
        if(metabolite_dict):
            result = {}
            for met in metabolite_dict:
                result[met.id] = met
            return result
        """
        if(input_dict):
            result={}
            for elm in input_dict:
                result[elm.id] = elm
            return result

    @property
    def ReactionsToGenesAssociations(self):
        rv = {}
        for reac in self.reactions:
            genes = self.GetGeneNames(reac.genes)
            rv[reac.id] = genes
        return rv

    @property
    def GenesToReactionsAssociations(self):
        rv = {}
        for gene in self.genes:
            reacs = self.GetReactionNames(gene.reactions)
            rv[gene.id] = reacs
        return rv

    @property
    def ReactionsToSubsystemsAssociations(self):
        rv = {}
        for reac in self.reactions:
            subsystems = getattr(reac, 'subsystem', '').split('|')
            if '' in subsystems:
                subsystems.remove('')
            rv[reac.id] = subsystems
        return rv

    @property
    def SubsystemsToReactionsAssociations(self):
        rv = defaultdict(list)
        reac2ss = self.ReactionsToSubsystemsAssociations
        for reac in reac2ss:
            for ss in reac2ss[reac]:
                rv[ss].append(reac)
        return dict(rv)

    @property
    def GenesToSubsystemsAssociations(self):
        rv = {}
        rs = self.ReactionsToSubsystemsAssociations
        for gene in self.genes:
            subsystems = set()
            reacs = self.GetReactionNames(gene.reactions)
            for reac in reacs:
                subsystems = subsystems.union(rs[reac])
            rv[gene.id] = list(subsystems)
        return rv

    @property
    def SubsystemsToGenesAssociations(self):
        rv = defaultdict(list)
        gene2ss = self.GenesToSubsystemsAssociations
        for gene in gene2ss:
            for ss in gene2ss[gene]:
                rv[ss].append(gene)
        return dict(rv)

    def NumberOfAssociations(self, associations='GeneReaction'):
        """ associations = GeneReaction|GeneSubsystem|ReactionSubsystem """
        if isinstance(associations, str):
            if ('gene' in associations.lower()) and ('reac' in associations.lower()):
                associations = self.GenesToReactionsAssociations
            elif ('gene' in associations.lower()) and ('subsystem' in associations.lower()):
                associations = self.GenesToSubsystemsAssociations
            elif ('reac' in associations.lower()) and ('subsystem' in associations.lower()):
                associations = self.ReactionsToSubsystemsAssociations
        rv = 0
        for b in associations:
            rv += len(associations[b])
        return rv


    ### SETTING BOUNDS, CONSTRAINTS AND OBJECTIVES ###############################

    ###### SETTING BOUNDS ###############################

    def SetBounds(self, bounds=1000.0, thres=1000.0):
        for reac in self.reactions:
            if reac.lower_bound <= -thres or reac.lower_bound==None:
                reac.lower_bound = -bounds
            if reac.upper_bound >= thres or reac.upper_bound==None:
                reac.upper_bound = bounds
        self.bounds = bounds

    def SetAllRev(self):
        """ remove all constraint!! set all reactions to reversible """
        for reac in self.reactions:
            reac.lower_bound = -self.bounds
            reac.upper_bound = self.bounds

    def SetMetsBounds(self, mets=None, direc="balance"):
        """ pre: direc="balance"|"out"|"in"|"free" """
        if not mets:
            mets = list(self.metabolites)
        elif isinstance(mets,str):
            mets = self.Metabolites(mets)
        for met in mets:
            if direc == "out":
                self.SetMetBounds(met, 0, None)
            elif direc == "in":
                self.SetMetBounds(met, None, 0)
            elif direc == "balance":
                self.SetMetBounds(met, 0, 0)
            elif direc == "free":
                self.SetMetBounds(met, None, None)

    def SetMetBounds(self, met, lo=0, hi=0):
        """ set metabolite mass balance constraint """
        met = self.GetMetabolite(met)
        reac = met.id + '_metbounds'
        if reac not in self.Reactions():
            self.AddReaction(reac,{met:-1},True)    # positive flux = export
        if lo == None:
            lo = -self.bounds
        if hi == None:
            hi = self.bounds
        if not numpy.allclose((lo,hi), (0,0)):
            self.SetConstraint(reac, lo, hi)
        else:
            self.DelReaction(reac)


    ###### SETTING CONSTRAINTS ###########################
    def GetConstraints(self, reaclist=None):
        if not reaclist:
            reaclist = self.Reactions()
        elif isinstance(reaclist,str):
            reaclist = self.Reactions(reaclist)
        rv = {}
        for reac in reaclist:
            rv[reac] = self.GetConstraint(reac)
        return rv

    def GetConstraint(self, reac):
        """ post: give constraints in forward direction """
        reac = self.GetReaction(reac)
        lb = reac.lower_bound
        ub = reac.upper_bound
        return (lb,ub)

    def SetConstraints(self, constraintdic):
        """ pre: {"R1":(lb,ub)} """
        for reac in constraintdic.keys():
            if (isinstance(constraintdic[reac],int) or isinstance(constraintdic[reac],float)):
                constraintdic[reac] = (constraintdic[reac],constraintdic[reac])
            self.SetConstraint(reac,constraintdic[reac][0],constraintdic[reac][1])

    def SetConstraint(self, reac, lb, ub=[]):
        """ pre: set constraint in forward direction """
        reac = self.GetReaction(reac)
        if (type(lb) == tuple) or (type(lb) == list):
            if len(lb) != 2:
                raise ValueError("length of constraint is not 2")
            else:
                if ub:
                    raise ValueError("wrong input for lb and ub")
                else:
                    lb, ub = lb
        if lb == None:
            lb = -self.bounds
        reac.lower_bound = lb
        if ub == None:
            ub = self.bounds
        reac.upper_bound = ub

    def SetFixedFlux(self, fluxdic):
        """ pre: fluxdic = {"R1":1} """
        for reac in fluxdic:
            self.SetConstraint(reac, fluxdic[reac], fluxdic[reac])

    def GetState(self):
        constraintdic = self.GetConstraints()
        obj = self.GetObjective(IncZeroes=True)
        objdirec = self.GetObjDirec()
        solver = self.solver
        quad = self.quadratic_component
        bounds = self.bounds
        sol = self.solution
        return {"constraints":constraintdic, "objective":obj,
                "objective_direction":objdirec, "solver":solver,
                "quaduatic_component":quad, "bounds":bounds, "solution":sol}

    def SetState(self, state, IncSol=True):
        self.SetConstraints(state["constraints"])
        self.SetObjective(state["objective"])
        self.SetObjDirec(state["objective_direction"])
        self.solver = state["solver"]
        self.quadratic_component = state["quaduatic_component"]
        self.SetBounds(bounds=state["bounds"])
        if IncSol:
            self.UpdateSolution(state["solution"])

    def SetSumReacsConstraint(self, reacsdic, bounds, name=None):
        """ bounds = val | (lb,ub) """
        if not name:
            name = str(reacsdic).replace(" ", "")
        if name+"_sum_reaction" in self.Reactions() or \
            name+"_sum_metabolite" in self.Metabolites():
            raise ValueError(name+" already in model, constraint not added")
        else:
            metabolite = Metabolite(id=name+"_sum_metabolite")
            for reac in reacsdic:
                reacval = reacsdic[reac]
                reac = self.GetReaction(reac)
                reac.add_metabolites({metabolite:reacval}) 
            self.AddReaction(name+"_sum_reaction", {metabolite:-1},
                             bounds=bounds)

    def DelSumReacsConstraint(self, sumreacs=None):
        if not sumreacs:
            sumreacs = []
            for reac in self.Reactions():
                if reac.endswith("_sum_reaction"):
                    sumreacs.append(reac)
            for met in self.Metabolites():
                if met.endswith("_sum_metabolite"):
                    sumreacs.append(met)
        elif isinstance(sumreacs,str):
            if not (sumreacs.endswith("_sum_reaction") or
                sumreacs.endswith("_sum_metabolite")):
                sumreacs = [sumreacs+"_sum_reaction",
                            sumreacs+"_sum_metabolite"]
        for sr in sumreacs:
            if sr in self.Reactions():
                self.DelReaction(sr)
            if sr in self.Metabolites():
                self.DelMetabolite(sr)

    def SetObjAsConstraint(self, name='Objective', subopt=1.0):
        self.Solve(PrintStatus=False)
        if self.Optimal():
            objval = self.GetObjVal()
            objective = self.GetObjective()
            if self.GetObjDirec == "minimize":
                bounds = (objval,objval/subopt)
            else:
                bounds = (subopt*objval,objval)
            self.SetSumReacsConstraint(reacsdic=objective, bounds=bounds,
                                       name=name)
        else:
            #self.GetStatusMsg()
            print("no solution from primary objective")

    def DelObjAsConstraint(self, name='Objective'):
        self.DelSumReacsConstraint(name)

    def SetReacsFixedRatio(self, ratiodic, GetMetName=False):
        """ ratiodic = {"R1":1,"R2":2} """
        temp_rd = {}
        for reac in ratiodic:
            temp_rd[self.GetReaction(reac)] = ratiodic[reac]
        reactions = self.GetReactions(temp_rd.keys())
        
        for reac in reactions[1:]:
            metname = reactions[0].id + "_" + reac.id + "_fixedratio"
            self.AddMetabolite(metname)
            met = self.GetMetabolite(metname)
            reactions[0].add_metabolites({met:-temp_rd[reac]})
            reac.add_metabolites({met:temp_rd[reactions[0]]})
        if GetMetName:
            return metname

    def DelReacsFixedRatio(self, fixedratio=None):
        if not fixedratio:
            self.DelMetabolites(self.Metabolites("_fixedratio"))
        elif isinstance(fixedratio, str):
            self.DelMetabolite(fixedratio)
        else:
            self.DelMetabolites(fixedratio)


    ###### SETTING OBJECTIVES ############################
    def SetObjDirec(self, direc="Min"):
        if direc in ["Min", "min", "minimize", "minimise"]:
            self.objective_direction = "minimize"
        elif direc in ["Max", "max", "maximize", "maximise"]:
            self.objective_direction = "maximize"
        else:
            raise exceptions.ValueError(direc)

    def SetObjective(self, objective):
        if isinstance(objective, dict):
            for reac in objective.keys():
                reacval = objective[reac]
                reac = self.GetReaction(reac)
                reac.objective_coefficient = reacval
        elif (isinstance(objective,str) or isinstance(objective, Reaction)):
            reac = self.GetReaction(objective)
            reac.objective_coefficient = 1
        else:   # list, tuple, set
            for reac in objective:
                reac = self.GetReaction(reac)
                reac.objective_coefficient = 1

    def SetQuadraticObjective(self, objective):
        if self.quadratic_component == None:
            diag = scipy.array([0]*len(self.reactions))
        else:
            diag = self.quadratic_component.diagonal()
        if isinstance(objective, dict):
            for reac in objective.keys():
                reacval = objective[reac]
                reac = self.GetReactionName(reac)
                idx = self.reactions.index(reac)
                diag[idx] = reacval
        elif (isinstance(objective, str) or isinstance(objective, Reaction)):
            reac = self.GetReactionName(objective)
            idx = self.reactions.index(reac)
            diag[idx] = 1
        else:   # list, tuple, set
            for reac in objective:
                reac = self.GetReactionName(reac)
                idx = self.reactions.index(reac)
                diag[idx] = 1
        q = scipy.sparse.diags([diag],[0]).todok()
        self.quadratic_component = q

    def ZeroObjective(self, IncQuad=True):
        for reac in self.Reactions():
            self.SetObjective({reac:0})
        if IncQuad:
            self.quadratic_component = None

    def GetObjVal(self):
        if self.solution != None: 
            return self.solution.objective_value
        else: 
            #print("no solution found")
            return None

    def GetObjDirec(self):
        return self.objective_direction

    def GetObjective(self, IncZeroes=False):
        obj = {}
        for reac in self.reactions:
            objcoef = reac.objective_coefficient
            if IncZeroes or (not numpy.allclose(objcoef,0.0)):
                obj[reac.id] = objcoef
        return obj

    #### SOLVING AND DISPLAYING SOLUTION ##################################
    
    ######## SOLVING #################################
    def Solve(self,PrintStatus=True, raise_error=False):
        sol = self.optimize(objective_sense=self.objective_direction, raise_error=raise_error)
        self.latest_solution = sol
        if PrintStatus:
            try: 
                print(self.solution.status)
            except AttributeError: 
                print("no solution")
        #print(sol.status)

    def MinFluxSolve(self, PrintStatus=True, PrimObjVal=True, norm="linear", 
                weighting='uniform', ExcReacs=[], adjusted=False, tol_step=1e-9,
                max_tol=1e-6, DisplayMsg=False, cobra=True, subopt=1.0):
        """ norm = "linear" | "euclidean"
            weighting = "uniform" | "random" """
        MinSolve.MinFluxSolve(self, PrintStatus=PrintStatus,
                              PrimObjVal=PrimObjVal, norm=norm,
                              weighting=weighting, ExcReacs=ExcReacs,
                              adjusted=adjusted, tol_step=tol_step, 
                              max_tol=max_tol, DisplayMsg=DisplayMsg, 
                              cobra=cobra, subopt=subopt)
        #return solfluxes

#    def AdjustedMinFluxSolve(self, PrintStatus=True, PrimObjVal=True, weighting='uniform', ExcReacs=[],
#                             SolverName=None, StartToleranceVal = 0,DisplayMsg=False):
#        
#        """ Adjusts the Minflux_objective constraint for feasible solution
#            StartToleranceVal = starting tolerance value"""
#        MinSolve.AdjustedMinFluxSolve(self, PrintStatus=PrintStatus,
#                              PrimObjVal=PrimObjVal,
#                              weighting=weighting, ExcReacs=ExcReacs,
#                                      SolverName=SolverName, Tolerance = #StartToleranceVal,DisplayMsg=DisplayMsg)

    def MinReactionsSolve(self, PrintStatus=True, PrimObjVal=True,
                          ExcReacs=[]):
        MinSolve.MinReactionsSolve(self, PrintStatus=PrintStatus,
                              PrimObjVal=PrimObjVal, ExcReacs=ExcReacs)

    def UpdateSolution(self,updated_sol):
        self.latest_solution = updated_sol

    @property
    def solution(self):
        return self.latest_solution
    #Changed solution.f to solution.objective_value because f was no longer an attribute
    def Optimal(self):
        if self.solution != None:
            if self.solution.status == "optimal" and not math.isnan(self.solution.objective_value):
                return True
            else: 
                return False
        else:
            #("no solution found")
            return False

    def GetStatusMsg(self):
        if self.solution != None: 
            return self.solution.status
        else: 
            #print("no solution")
            return "no solution"

    ######## DISPLAYING SOLUTIONS #################################################
    def GetSol(self, IncZeroes=False, AsMtx=False, sol=None, FixSumReacs=True,
               FixMetBounds=True, f=None, met=None, reacs=None, AsID=False,
               tol=1e-10):
        if not sol:
            #sol = flux(self.solution.x_dict
            #        ) if dict(self.solution.x_dict) != None else flux()
            if self.solution.status == 'optimal':
                #sol_object = Reversible.MergeSolution(self.solution)
                #sol = flux(sol_object.fluxes.to_dict())
                sol = flux(self.solution.fluxes.to_dict())
            else: 
                print("no solution found")
                sol = flux({})  
                return sol
        else:
            sol = flux(sol)
        if IncZeroes:
            for reac in self.Reactions():
                if reac not in sol.keys():
                    sol[reac] = 0.0
                else:
                    if abs(sol[reac]) < tol:
                        sol[reac] = 0.0
        else:
            for reac in list(sol.keys()):
                if abs(sol[reac]) < tol :
                    del sol[reac]
        if FixSumReacs:
            for reac in list(sol.keys()):
                if reac.endswith("_sum_reaction"):
                    del sol[reac]
        if FixMetBounds:
            for reac in list(sol.keys()):
                if reac.endswith("_metbounds"):
                    del sol[reac]
        if f != None:
            for reac in list(sol.keys()):
                if f not in reac:
                    del sol[reac]
        if met != None:
            for reac in list(sol.keys()):
                if self.GetMetabolite(met) not in self.InvolvedWith(reac):
                    del sol[reac]
        if reacs != None:
            reacs = self.GetReactionNames(reacs)
            for reac in list(sol.keys()):
                if reac not in reacs:
                    del sol[reac]
        if AsID: 
            newsol = {}
            for reac in sol.keys():
                solval = sol[reac]
                reac = self.GetReaction(reac)
                newsol[reac] = solval
            sol = newsol
        if AsMtx:
            rv = sol.AsMtx()
        else:
            rv = sol
        return rv

    def PrintSol(self, lo=0, hi=float('inf'), f=None, sol=None, met=None,
                 reacs=None, Sort="value", IncZeroes=False, sortabs=True,
                 reverse=True, FixMetBounds=True):
        if not sol:
            sol = self.GetSol(IncZeroes=IncZeroes, AsMtx=False,
                              FixMetBounds=FixMetBounds, f=f, met=met,
                              reacs=reacs)
        flux(sol).Print(lo=lo, hi=hi, f=f, Sort=Sort, sortabs=sortabs,
                  reverse=reverse)

    def SolsDiff(self, sol1, sol2, IncZeroes=False, AsMtx=False, tol=1e-10):
        return flux(sol1).Diff(sol2, IncZeroes=IncZeroes, AsMtx=AsMtx, tol=tol)

    def NetStoi(self, sol=None, IncZeroes=False, tol=1e-10):
        if sol == None:
            sol = self.GetSol()
        rv = {}
        for reac in sol:
            reacstoi = self.InvolvedWith(reac)
            for met in reacstoi:
                if met in rv.keys():
                    rv[met] += reacstoi[met]*sol[reac]
                else:
                    rv[met] = reacstoi[met]*sol[reac]
        if not IncZeroes:
            for met in list(rv.keys()):
                if abs(rv[met]) < tol:
                    del rv[met]
        return rv

    def ProduceMetabolites(self, mets=None, indep=False, rc="all"):
        """ pre: rv="all"|"Produce"|"Not Produce", indep=met produced independently
           post: returns dic or list of metabolites that can(not) be produced given defined inputs """
        state = self.GetState()
        self.ZeroObjective()
        self.SetObjDirec("Min")
        prod = []
        notprod = []
        if not indep:
            self.SetMetsBounds(direc="out")
        if not mets:
            mets = list(self.Metabolites())
        for met in mets:
            self.SetMetBounds(met, 1, 1)
            self.Solve(False)
            if self.Optimal():
                prod.append(met)
            #elif self.GetStatusMsg() == "infeasible":
            #    notprod.append(met)
            else:
                notprod.append(met)
                #print("error")
            if indep:
                self.SetMetBounds(met, 0, 0)
            else:
                self.SetMetBounds(met, 0, None)
        self.SetState(state)
        self.SetMetsBounds(direc="balance")
        rv = {"all":{"Produce":prod, "Not Produce":notprod}, "Produce":prod,
                                                      "Not Produce":notprod}
        return rv[rc]

    def BlockedMetabolites(self, metabolites=None, fva=None, tol=1e-10):
        if not metabolites:
            metabolites = self.Metabolites
        if fva == None:
            fva = self.FVA()
        allowedreacs = fva.Allowed(tol=tol)
        allowed_mets = []
        for r in allowedreacs:
            involved_mets = self.InvolvedWith(r)
            allowed_mets = list(set(allowed_mets).union(involved_mets))
        rv = self.GetMetaboliteNames(list(set(
                self.metabolites).difference(allowed_mets)))
        return rv



    ### FVA, FCA AND PARETO ANALYSIS #########################################

    ####### FVA ########################################
    def FVA(self, reaclist=None, subopt=1.0, IncZeroes=True, VaryOnly=False,
            AsMtx=False, tol=1e-10, PrintStatus=False, cobra=True,
            processes=None, loopless=False, pfba_factor=None, reset_state=True):
        rv = FVA.FVA(self, reaclist=reaclist, subopt=subopt,
            IncZeroes=IncZeroes, VaryOnly=VaryOnly, AsMtx=AsMtx, tol=tol,
            PrintStatus=PrintStatus, cobra=cobra, processes=processes,
            loopless=loopless, pfba_factor=pfba_factor,reset_state=reset_state) 
        return rv

    def MinFluxFVA(self, reaclist=None, subopt=1.0, IncZeroes=True, 
                   VaryOnly=False, AsMtx=False, tol=1e-10, PrintStatus=False, 
                   cobra=True, processes=None, weighting='uniform', ExcReacs=[],
                   loopless=False, pfba_factor=1.0, reset_state=True):
        rv = FVA.MinFluxFVA(self, reaclist=reaclist, subopt=subopt,
            IncZeroes=IncZeroes, VaryOnly=VaryOnly, AsMtx=AsMtx, tol=tol,
            PrintStatus=PrintStatus, cobra=cobra, processes=processes,
            weighting=weighting, ExcReacs=ExcReacs, 
            loopless=loopless, pfba_factor=pfba_factor, reset_state=reset_state)
        return rv
            
    def AllFluxRange(self, tol=1e-10, processes=None, reset_state=True):
        return FVA.AllFluxRange(self, tol=tol, processes=processes, reset_state=reset_state)

    def FluxRange(self, obj, tol=1e-10, reset_state=True, return_reac=False):
        """ post: changes objective if resetobj = False!!! """
        return FVA.FluxRange(self, obj=obj, tol=tol, reset_state=reset_state, return_reac=return_reac)

    def FluxVariability(self, reffva    =None, fva=None, excreacs=[], tol=1e-10,
                        getratio=False):
        rv = FVA.FluxVariability(self, reffva=reffva, fva=fva,
            excreacs=excreacs, tol=tol, getratio=getratio)
        return rv

    def InternalCycles(self, allowedreacs=None, reacsbounds={}, tol=1e-10):
        """ pre: reacsbounds={reac:(lo,hi)}, all external reactions blocked
           post: model with reactions in internal cycles (for doing elementary modes) """
        rv = FVA.InternalCycles(self, allowedreacs=allowedreacs,
                            reacsbounds=reacsbounds, tol=tol)
        return rv

    ###### FCA ########################################################
    def FCA(self, reacs=None, rangedict=None, tol=1e-10):
        """ pre: lp object
           post: returns dataset of coupling """
        return FCA.FCA(self, reacs=reacs, rangedict=rangedict, tol=tol)

    ###### PARETO ANALYSIS ##############################################################
    def Pareto(self, objectives, objdirec, runs, GetPoints=True, tol=1e-10):
        """ pre: objective = [["reac"],{"reac2":x}]
           post: turning points of Pareto front """
        return Pareto.Pareto(self, objectives, objdirec, runs,
                                GetPoints=GetPoints, tol=tol)


    ### SCANS ###############################################################
    def ConstraintScan(self, cd, lo, hi, n_p, MinFlux=True, IncZeroes=True,cobra=True):
        """ scan one reaction flux
            pre: cd = sum of reaction fluxes dictionary """
        return Scan.ConstraintScan(self, cd, lo, hi, n_p, MinFlux=MinFlux, IncZeroes=IncZeroes,cobra=cobra)

    def RatioScan(self, reac1, reac2, n_p, lo=0, hi=1, flux_val=None,
                  IncZeroes=True, rev=False):
        """ scan the ratio of two reaction fluxes
            pre: flux_val = a fixed flux for the sum of the two reactions """
        return Scan.RatioScan(self, reac1, reac2, n_p, lo=lo, hi=hi,
             flux_val=flux_val, IncZeroes=IncZeroes, rev=rev)

    def Constraint2DScan(self, cd1, lo1, hi1, cd2, lo2, hi2, n_p,
                        IncZeroes=True):
        """ scan two reaction fluxes simultaneously """
        return Scan.Constraint2DScan(self, cd1, lo1, hi1, cd2, lo2, hi2, n_p,
                                        IncZeroes=IncZeroes)

    def ConstraintRandomMinFluxScan(self, cd, lo, hi, n_p, it, IncZeroes=True,
                                    reacs=None, exc=[], processes=None):
        """ same as ConstraintScan except using RWFM rather than FBA
            pre: cd = sum of reaction fluxes dictionary """
        return Scan.ConstraintRandomMinFluxScan(self, cd, lo, hi, n_p, it,
             IncZeroes=IncZeroes, reacs=reacs, exc=exc)

    def RatioRandomMinFluxScan(self, reac1, reac2, n_p, it, lo=0, hi=1,
                flux_val=None, IncZeroes=True, reacs=None, exc=[], rev=False,
                processes=None):
        """ same as RatioScan except using RWFM rather than FBA """
        return Scan.RatioRandomMinFluxScan(self, reac1, reac2, n_p, it, lo=lo,
            hi=1, flux_val=flux_val, IncZeroes=IncZeroes, reacs=reacs, exc=exc, rev=rev)

    def WeightingScan(self, objdic, lo, hi, n_p):
        """ scan by changing the objective coefficients for a subset of
            reactions in the objective function """
        return Scan.WeightingScan(self, objdic, lo, hi, n_p)

    def MatchFlux(self,md,vd,lo=0,hi=100,samedirec=True,count=50,display=False,tol=1e-6):
        """ Not functional! To be modified in .analysis.Match
            pre: md = matched flux dict; vd = vary flux dict / vary met bound dict """
        pass

    def MatchRatio(self,numdic,domdic,val,vd,lo=0,hi=100,samedirec=True,count=50,display=False,tol=1e-6):
        """ Not functional! To be modified in .analysis.Match """
        pass

    def MatchScan(self,cd,clo,chi,md,vd,vlo,vhi,n_p,samedirec=True,count=50,display=False,tol=1e-6,IncZeroes=True):
        """ Not functional! To be modified in .analysis.Scan """
        pass


    def PhasePlane(self, reac1, reac2, reac1_max=20, reac2_max=20, reac1_n=50,
            reac2_n=50, solver=None, n_processes=1, tol=1e-10):
        pass
        """Not functional, calculate_phenotype_phase_plane not found in cobra module"""
        """return phenotype_phase_plane.calculate_phenotype_phase_plane(self,
            reac1=reac1 ,reac2=reac2 , reac1_max=reac1_max,
            reac2_max=reac2_max, reac1_n=reac1_n, reac2_n=reac2_n,
            solver=solver, n_processes=n_processes, tol=tol)
        """


    #### MOMA, ROOM, GEOMETRIC SOL, FLUX RANGE, FLUX SUM FUNCTIONS ###################

    ######## MOMA #############################################
    def LinearMOMA(self, refflux):
        MOMA.LinearMOMA(self, refflux=refflux)

    def MinDiffFromFlux(self, fluxdist, reacs=None, it=0, cleanup=True):
        """ pre: fluxdist = {reac:flux_val} """
        MOMA.MinDiffFromFlux(self, fluxdist=fluxdist, reacs=reacs, it=it,
                                cleanup=cleanup)

    def CleanUpTempVar(self, var):
        MOMA.CleanUpTempVar(self, var=var)

    def MOMA(self,refflux):
        MOMA.MOMA(self, refflux=refflux)

    def MOMA2mutant(self, objective_sense='maximize', solver=None, tolerance_optimality=1e-8, tolerance_feasibility=1e-8,minimize_norm=False, the_problem='return', lp_method=0,combined_model=None, norm_type='euclidean'):
        moma.add_moma(self)

    ######## ROOM ############################################
    def ROOM(self, refflux, reactions=None, delta=0, tol=0, IncZeroes=False,
             AsMtx=False, f=None, reset_state=True):
        """ refflux = {reac_name:flux_val} """
        ROOM.ROOM(self, refflux=refflux, reactions=reactions,
                         delta=delta, tol=tol, IncZeroes=IncZeroes,
                         AsMtx=AsMtx, f=f, reset_state=reset_state)

    def GeometricSol(self, IncZeroes=True, AsMtx=False, tol=1e-6, Print=False):
        return GeometricFBA.GeometricSol(self, IncZeroes=IncZeroes,
                        AsMtx=AsMtx, tol=tol, Print=Print)

    ######## FLUX RANGE #######################################
    def RandomMinFlux(self, it=1, reacs=None, exc=[], processes=None):
        return RWFM.RandomMinFlux(self, it=it, reacs=reacs, exc=exc,
                            processes=processes)

    def FluxRangeDiff(self, fd1, fd2, thres=1):
        return fd1.FluxRangeDiff(fd2, thres=thres)

    def FluxDiffDirec(self, fd1, fd2):
        return fd1.FluxDiffDirec(fd2)

    def FluxRangeOverlap(self, fd1, fd2):
        return fd1.FluxRangeOverlap(fd2)

    ######## FLUX SUM #######################################################
    def FluxSum(self, met, tol=1e-10):
        rv = FluxSum.FluxSum(self, met=met, tol=tol)
        return rv

    def ProducedBy(self, met, FixBack=True):
        return FluxSum.ProducedBy(self, met=met, FixBack=FixBack)

    def ConsumedBy(self,met,FixBack=True):
        return FluxSum.ConsumedBy(self, met=met, FixBack=FixBack)

    #### GRAPH FUNCTIONS ###########################################################

    def DeadMetabolites(self, fva=None):
        """ metabolites not involved in any allowed reactions """
        return Graph.DeadMetabolites(self, fva=fva)

    def NoDeadEndModel(self):
        return Graph.NoDeadEndModel(self)

    def DeadEndMetabolites(self, nodeadendmodel=None):
        """ iteratively identify peripheral metabolites """
        return Graph.DeadEndMetabolites(self, nodeadendmodel=nodeadendmodel)

    def PeripheralMetabolites(self, rc="all"):
        """ pre: rv ="all"|"Produced"|"Consumed"|"Orphan"
           post: returns dic or list of deadend metabolites """
        return Graph.PeripheralMetabolites(self, rc=rc)

    def ChokepointReactions(self):
        return Graph.ChokepointReactions(self)

    def GetNeighbours(self, name, exclude=[]):
        return Graph.GetNeighbours(self, name=name, exclude=exclude)

    def GetNeighboursAsDic(self, name, exclude=[]):
        """  pre: name and exclude are reactions or metabolites names
            post: {iw:[neighbours]} """
        return Graph.GetNeighboursAsDic(self, name=name, exclude=exclude)

    def Degree(self, name, bipartite=True):
        return Graph.Degree(self, name, bipartite=bipartite)

    def DegreeDist(self, node_type="metabolites", bipartite=True):
        """ node_type = "metabolites" | "reactions" """
        return Graph.DegreeDist(self, node_type="metabolites", bipartite=True)

    def MetabolitesDegree(self, mets=None, bipartite=True):
        if(not isinstance(mets,list)):
            mets=[mets]
        return Graph.MetabolitesDegree(self, mets=mets, bipartite=bipartite)

    def ReactionsDegree(self, reacs=None, bipartite=True):
        return Graph.ReactionsDegree(self, reacs=None, bipartite=True)


    ## MODEL COMPARISON FUNCTIONS ####################################################
    def CompareModel(self, m2): 
        """
        params self,m2: two model objects to be compared 
        
        returns comparison(dict):
         comparison["reactions"][1] contains a list of reactions unique to only m1 
         comparison["reactions"][2] contains a list of reactions unique to only m2 
         comparison["reactions"][3] contains a list of reactions in both m1 and m2
         comparison["metabolites"][1] contains a list of reactions unique to only m1 
         comparison["metabolites"][2] contains a list of reactions unique to only m2 
         comparison["metabolites"][3] contains a list of reactions in both m1 and m2
         comparison["genes"][1] contains a list of reactions unique to only m1 
         comparison["genes"][2] contains a list of reactions unique to only m2 
         comparison["genes"][3] contains a list of reactions in both m1 and m2
        """
        comparison = {}
        comparison["reactions"] = []
        comparison["metabolites"] = []
        comparison["genes"] = []
    
        comparison["reactions"].append(set(self.Reactions()).difference(set(m2.Reactions())))
        comparison["reactions"].append(set(m2.Reactions()).difference(set(self.Reactions())))
        comparison["reactions"].append(set(self.Reactions()).intersection(set(m2.Reactions())))

        comparison["metabolites"].append(set(self.Metabolites()).difference(set(m2.Metabolites())))
        comparison["metabolites"].append(set(m2.Metabolites()).difference(set(self.Metabolites())))
        comparison["metabolites"].append(set(self.Metabolites()).intersection(set(m2.Metabolites())))
    
        comparison["genes"].append(set(self.Genes()).difference(set(m2.Genes())))
        comparison["genes"].append(set(m2.Genes()).difference(set(self.Genes())))
        comparison["genes"].append(set(self.Genes()).intersection(set(m2.Genes())))
    
        return comparison 
    



import random
try:
    import scipy
except ImportError:
    pass
from . import ROOM
from cobra.exceptions import Infeasible
#from cobra.manipulation import modify
from ..manipulation import Reversible
from cobra.flux_analysis.parsimonious import pfba

#def SetValAsConstraint(model, name,objval,objective): # Not require to solve here again
#        bounds = (objval,objval)
#        model.SetSumReacsConstraint(reacsdic=objective, bounds=bounds,name=name)


def SetLinearMinFluxObjective(model, weighting='uniform', ExcReacs=[]):
    """ assume reversible reactions are split """
    ExcReacs = model.GetReactionNames(ExcReacs)
    for reaction in model.reactions:
        if not (reaction.id.endswith("_sum_reaction") or
                reaction.id.endswith("_metbounds") or
                (reaction.id.split('_reverse')[0] in ExcReacs)):
            if weighting == 'uniform':
                reaction.objective_coefficient = 1
            elif weighting == 'random':
                reaction.objective_coefficient = random.random()
            else:
                #print "wrong weighting"
                raise NameError(weighting)
    model.SetObjDirec("Min")
    

def MinFluxSolve(model, PrintStatus=True, PrimObjVal=True, norm="linear", 
                weighting='uniform', ExcReacs=[], adjusted=False, tol_step=1e-9,
                max_tol=1e-8, DisplayMsg=True, cobra=True, subopt=1.0):
    """ norm = "linear" | "euclidean"
        weighting = "uniform" | "random" """
    
#    """Temporary fix, need to change the structure of saving solution objects"""
    if norm == "linear" and weighting=="uniform" and (not ExcReacs) and (not adjusted) and cobra:
#            """Temporary fix, need to change the structure of saving solution objects"""
#            from cobra.flux_analysis.parsimonious import pfba
        sol=None
        try:
            sol = pfba(model, fraction_of_optimum=subopt)
        except Infeasible:
            if DisplayMsg:
                print("no solution")
            model.UpdateSolution(None)
            return
        model.UpdateSolution(sol)
        if DisplayMsg:
            try: 
                print(sol.status)
            except AttributeError: 
                print("no solution")

        return
#            return sol.fluxes
#
#    RemoveReverse(model)
    model.Solve(PrintStatus=PrintStatus)
    if model.Optimal():
        state = model.GetState()
        objective = model.GetObjective()
        objval = model.GetObjVal()
        objbounds = (objval - abs(1-subopt)*objval, objval + abs(1-subopt)*objval)
        #SetValAsConstraint(model,name="MinFlux_Objective", objval=objval,objective=objective)
        #model.SetSumReacsConstraint(reacsdic=objective, bounds=objval,name="MinFlux_Objective")
        model.SetSumReacsConstraint(reacsdic=objective, bounds=objbounds,name="MinFlux_Objective")
        if norm == "linear":
            #modify.convert_to_irreversible(model)
            model.SplitRev()
            SetLinearMinFluxObjective(model=model, weighting=weighting, ExcReacs=ExcReacs)
#            ExcReacs = model.GetReactionNames(ExcReacs)
#            for reaction in model.reactions:
#                if not (reaction.id.endswith("_sum_reaction") or
#                        reaction.id.endswith("_metbounds") or
#                        (reaction.id.split('_reverse')[0] in ExcReacs)):
#                    if weighting == 'uniform':
#                        reaction.objective_coefficient = 1
#                    elif weighting == 'random':
#                        reaction.objective_coefficient = random.random()
#                    else:
#                        #print "wrong weighting"
#                        raise NameError(weighting)
           
            model.SetObjDirec("Min")
            model.Solve(PrintStatus=DisplayMsg)
            
            if adjusted:
                Tolerance = 0
                while (model.Optimal()==False and Tolerance < max_tol):
                    Tolerance = Tolerance + tol_step
                    model.DelObjAsConstraint("MinFlux_Objective")
                    bounds = (objval-Tolerance, objval+Tolerance)
                    if DisplayMsg:
                        print('Objective value = ' + str(objval))
                        print("Modified 'MinFlux_Objective' bounds = " + str(bounds))
                    model.SetSumReacsConstraint(reacsdic=objective, bounds=bounds, name="MinFlux_Objective")
                    model.SetObjDirec("Min")
                    model.Solve(PrintStatus=DisplayMsg)
            #modify.revert_to_reversible(model)
            model.MergeRev(True)
            #print(model.GetConstraints())

        if norm == "euclidean":
            num_reacs = len(model.reactions)
            model.quadratic_component = scipy.sparse.identity(num_reacs).todok()
            model.SetObjDirec("Min")
            model.Solve(PrintStatus=DisplayMsg)

#"""This whole section used to be run after, either linear or euclidean norms, but tabbed to accomodate temporary change described above"""

        model.DelObjAsConstraint("MinFlux_Objective")
        model.SetState(state, IncSol=False)
        if PrimObjVal:
            try: 
                model.solution.objective_value = state["solution"].objective_value
            except AttributeError: 
                pass
        #return model.GetSol()
    else:
        print("not feasible")
        #return model.GetSol()


#def RevSolve(model,objective,objval,Tolerance,DisplayMsg=False):
#    #RemoveReverse(model)
#    model.DelObjAsConstraint("MinFlux_Objective")    
#    bounds = (objval - Tolerance , objval + Tolerance)
#    if DisplayMsg:
#        print 'Objective value = ',objval
#        print "Modified 'MinFlux_Objective' bounds = ",bounds
#    model.SetSumReacsConstraint(reacsdic=objective, bounds=bounds,name="MinFlux_Objective")
#    model.SetObjDirec("Min")
#    model.Solve(PrintStatus=True)

#def AdjustedMinFluxSolve(model,PrintStatus=True, PrimObjVal=True, weighting='uniform', ExcReacs=[], Tolerance=0, DisplayMsg=False):
##SolverName=None
#    """ weighting = "uniform" | "random"
#        Only enters to the adjusted 'ObjVal' mode if solution after SplitRev() == infeasible"""
#    #RemoveReverse(model)
#    model.Solve(PrintStatus=PrintStatus)
#
#    if model.Optimal():
#        state = model.GetState()
#        objective = model.GetObjective()
#        objval = model.GetObjVal()
#        model.SetSumReacsConstraint(reacsdic=objective, bounds=objval,name="MinFlux_Objective")
#        #SetValAsConstraint(model,name="MinFlux_Objective",objval=objval,objective=objective)
#        #modify.convert_to_irreversible(model)
#        model.SplitRev()
#        ExcReacs = model.GetReactionNames(ExcReacs)
#        for reaction in model.reactions:
#            if not (reaction.id.endswith("_sum_reaction") or  reaction.id.endswith("_metbounds") or (reaction.id.split('_reverse')[0] in ExcReacs)):
#                if weighting == 'uniform':
#                    reaction.objective_coefficient = 1
#                elif weighting == 'random':
#                        reaction.objective_coefficient = random.random()
#                else:
#                    raise NameError(weighting)
#                    
#        model.SetObjDirec("Min")
##        if SolverName!=None:
##            model.solver = SolverName
#            
#        model.Solve(PrintStatus=False)
#            
#        while (model.GetStatusMsg()=='infeasible' or Tolerance >= 1e-5):
#            Tolerance = Tolerance + 1e-9
#            RevSolve(model,objective=objective,objval=objval,Tolerance=Tolerance,DisplayMsg=DisplayMsg)
#                
#        model.MergeRev(True)
#
#        model.DelObjAsConstraint("MinFlux_Objective")
#        model.SetState(state, IncSol=False)
#        if PrimObjVal:
#            try:
#            	model.solution.objective_value = state["solution"].objective_value
#            except AttributeError: 
#            	pass

def MinReactionsSolve(model, PrintStatus=True, PrimObjVal=True, ExcReacs=[]):
    #RemoveReverse(model)
    model.Solve(PrintStatus=PrintStatus)
    if model.Optimal():
        state = model.GetState()
        model.SetObjAsConstraint("MinReacs_Objective")
        ROOM.ROOM(model, {}, reset_state=False)
        model.DelObjAsConstraint("MinReacs_Objective")
        model.SetState(state, IncSol=False)
        if PrimObjVal:
            try:
                model.solution.objective_value = state["solution"].ibjective_value
            except AttributeError: 
                pass

#def RemoveReverse(model): 
#    reverse_reactions = [x for x in model.reactions
#                         if x.id.endswith('_reverse')]
#    model.remove_reactions(reverse_reactions)
#    return 



from ..classes.fca import fca

def FCA(model, reacs=None, rangedict=None, tol=1e-10):
    """ pre: model object
       post: returns dataset of coupling """
    state = model.GetState()

    if rangedict == None:
        rangedict = model.AllFluxRange(tol=tol)
#        allreacs = self.Reactions()   # Forward reaction names
#        allowed = rangedict.Allowed()
#        blockedreacs = rangedict.Blocked().keys()
    allowedreacs = rangedict.Allowed().keys()
    #blocked = self.BlockedReactions(rangedict=rangedict,tol=tol) # Blocked Reactions
    #allreacs = list(set(allreacs).difference(blocked))     # Allowed reactions (forward names only)
    backwardonly = rangedict.BackwardOnly().keys()
    #rev = self.ReversedReaction(rangedict=rangedict,tol=tol)   # Reversible reacs only in -ve direction
    forwardonly = rangedict.ForwardOnly().keys()
    bothdirec = rangedict.BothDirections().keys()
#        for reac in forwardonly:
#            self.SetConstraint(reac,0,None)     # Remove fixed constraints, only looking at structual properties
#        for reac in backwardonly:
#            self.SetConstraint(reac,None,0)
#        for reac in bothdirec:
#            self.SetConstraint(reac,None,None)
    constraintdic = model.GetConstraints()
    ds = fca(columns=allowedreacs)
    if not reacs:
#            reacs = allreacs
        reacs = allowedreacs
    for reac in reacs:          # compute whole matrix for given reactions
        if reac in allowedreacs:                    # reac not blocked
            if reac in backwardonly:
                if rangedict[reac][0]==float("-inf") or rangedict[reac][0]==None:
                    model.SetFixedFlux({reac:rangedict[reac][1]-1})  # any negative number within the range
                else:
                    model.SetFixedFlux({reac:sum(rangedict[reac])/2.0})
            elif reac in forwardonly:
                if rangedict[reac][1]==float("inf") or rangedict[reac][1]==None:
                    model.SetFixedFlux({reac:rangedict[reac][0]+1})  # any positive number within the range
                else:
                    model.SetFixedFlux({reac:sum(rangedict[reac])/2.0})
            elif reac in bothdirec:
                model.SetFixedFlux({reac:rangedict[reac][1]/2.0})  # any positive number within range
            ds.loc[reac] = 0
            for reac2 in allowedreacs:
                if reac != reac2:
                    lo,hi = model.FluxRange(reac2)
                    if reac in backwardonly:
                        ds.loc[reac,reac2] = ds.RangeType(-hi, -lo, tol,
                                                    bounds=model.bounds)
                    else:
                        ds.loc[reac,reac2] = ds.RangeType(lo, hi, tol,
                                                    bounds=model.bounds)
            lb,ub = constraintdic[reac]
            model.SetConstraint(reac, lb, ub)
    model.SetState(state)
    return fca(ds)



import pandas

def WriteReactionsToMetabolitesNetwork(model, filename, ExtReacs=[], ExtMets=[]):
    """ ExtReacs = []; ExtMets = [reac_id] """
    rm = pandas.DataFrame(columns=['reaction','metabolite','stoichiometry'])
    for reac in model.reactions:
        if reac.id not in ExtReacs:
            mets = reac.metabolites
            for met in mets:
                if met.id not in ExtMets:
                    rm = rm.append({'reaction':reac.id, 'metabolite':met.id, 'stoichiometry':reac.metabolites[met]}, ignore_index=True)
    if filename.endswith('xls') or filename.endswith('xlsx'):
        rm.to_excel(filename, index=False)
    else:
        rm.to_csv(filename, index=False, sep='\t')

def WriteReactionsToReactionsNetwork(model, filename, ExtReacs=[], ExtMets=[]):
    """ ExtReacs = []; ExtMets = [reac_id] """
    rr = pandas.DataFrame(columns=['reaction1','reaction2'])
    reacs_mets = {}
    for reac in model.reactions:
        reacs_mets[reac] = set(reac.metabolites)
    for r1 in range(len(model.reactions)):
        reac1 = model.reactions[r1]
        for r2 in range(r1,len(model.reactions)):
            reac2 = model.reactions[r2]
            if len(reacs_mets[reac1].intersection(reacs_mets[reac2])) > 0:
                rr = rr.append({'reaction1':reac1.id,
                                'reaction2':reac2.id},
                                ignore_index=True)
    if filename.endswith('xls') or filename.endswith('xlsx'):
        rr.to_excel(filename, index=False)
    else:
        rr.to_csv(filename, index=False, sep='\t')

def WriteMetabolitesToMetabolitesNetwork(model, filename, ExtReacs=[], ExtMets=[]):
    """ ExtReacs = []; ExtMets = [reac_id] """
    mm = pandas.DataFrame(columns=['metabolite1','metabolite2'])
    mets_reacs = {}
    for met in model.metabolites:
        mets_reacs[met] = set(met.reactions)
    for m1 in range(len(model.metabolites)):
        met1 = model.metabolites[m1]
        for m2 in range(m1,len(model.metabolites)):
            met2 = model.metabolites[m2]
            if len(mets_reacs[met1].intersection(mets_reacs[met2])) > 0:
                mm = mm.append({'metabolite1':met1.id,
                                'metabolite2':met2.id},
                                ignore_index=True)
    if filename.endswith('xls') or filename.endswith('xlsx'):
        mm.to_excel(filename, index=False)
    else:
        mm.to_csv(filename, index=False, sep='\t')

def WriteReactionsAttributes(model, filename, attributes=[]):
    df = pandas.DataFrame(columns=attributes)
    df.index.name = 'ID'
    for reac in model.reactions:
        for attr in attributes:
            df.loc[reac.id,attr] = getattr(reac,attr)
    if filename.endswith('xls') or filename.endswith('xlsx'):
        df.to_excel(filename)
    else:
        df.to_csv(filename, sep='\t')

def WriteMetabolitesAttributes(model, filename, attributes=[]):
    df = pandas.DataFrame(columns=attributes)
    df.index.name = 'ID'
    for met in model.metabolites:
        for attr in attributes:
            df.loc[met.id,attr] = getattr(met,attr)
    if filename.endswith('xls') or filename.endswith('xlsx'):
        df.to_excel(filename)
    else:
        df.to_csv(filename, sep='\t')



import setuptools
import os
thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'
install_requires = [] # Better to write it in requirements.txt
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        install_requires = f.read().splitlines()

with open ("README.md","r") as fh:
    long_description=fh.read()

setuptools.setup(
    name="scobra",
    version="0.1.3",
    license='MIT',
    author="Maurice Cheung",
    author_email="maurice.cheung@yale-nus.edu.sg",
    description="Supplementary package for COBRApy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mauriceccy/scobra",
    packages=setuptools.find_packages(),
    install_requires=install_requires,
    classifiers=[ #find list of classifiers here https://pypi.org/classifiers/
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Development Status :: 5 - Production/Stable',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Science/Research',      # Define your audience
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'License :: OSI Approved :: MIT License',   # license
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',     
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ],
    python_requires='>=3.5',
)



from .matrix import matrix
import scipy

class pareto(matrix):

    def __init__(self,*args,**kwargs):
        super(pareto,self).__init__(*args,**kwargs)


    def GetParetoPoints(self,tol=1e-10):
        po2 = pareto()
        for r in self.columns.tolist():
            if r.startswith('Obj'):
                po2[r] = self[r]
        pp = [po2.iloc[0,:].tolist()]
        rv = pareto(columns=po2.columns)
        index = rv.index.tolist()
        rv = rv.append(po2.iloc[0,:])
        rv.index = index + [po2.index.tolist()[0]] 
        for row in po2.index.tolist():
            po2r = po2.loc[row]
            diff = True
            for p in pp:
                if scipy.spatial.distance.cityblock(p,po2r) > tol:
                    diff = False
            if diff:
                index = rv.index.tolist()
                rv = rv.append(po2r)
                rv.index = index + [row]
                pp.append(po2r)
        return rv

